﻿#ifndef _MIR_DATABASEMODULE_H_
#define _MIR_DATABASEMODULE_H_
/*
 * Copyright (c) 2011,
 * Technical University Of Gdansk,
 * Faculty of Electronics, Telecommunications and Informatics,
 * Multimedia Systems Department
 * and/or its affiliates. All rights reserved.
 */

#include <boost/regex.hpp>
#include <boost/ptr_container/ptr_vector.hpp>

#include <string>

#include <ibpp.h>

#include "mir_exceptions.h"
#include "mir_databasetypes.h"

#include <unicode/unistr.h>
#include <unicode/regex.h>
#include "utils/mir_enums.h"



namespace SyNaT
{
  namespace MIR
  {
    using namespace std;
    using namespace IBPP;

    /**
     * A <code>DatabaseModule</code> provides different database centric
     * functions like:
     * -# insertIntoTable(...)
     * -# updateTable(...)
     * -# deleteFromTable(...)
     * and many others.
     *
     * @author  Łukasz Matuszewski
     * @version $Id: mir_databasemodule.h 399 2011-09-08 09:40:52Z lmatusz $
     */
    class DatabaseModule
    {
      protected:
        Database                   m_db;

        int64_t                    m_adminID;
        int64_t                    m_loggedUserId;
        int64_t                    m_userPrivilages;
        string                     m_name;
        string                     m_surname;
        string                     m_hashpass;

        bool                       m_dbAssigned;

      public:


        int64_t                    getNextID(const UnicodeString &tablePrefix, Transaction *tr);
        int64_t                    getNextID(const string &tablePrefix, Transaction *tr);

        DatabaseModule();
        DatabaseModule(const UnicodeString &serverName, const UnicodeString &databaseName,
                                    const UnicodeString &userName, const UnicodeString &userPassword,
                                    const UnicodeString &roleName, const UnicodeString &charSet,
                                    const UnicodeString &createParams, const UnicodeString &login, const UnicodeString &loginPassword, bool isSha1 = true); // throw(...);

        DatabaseModule(const string &serverName, const string &databaseName, 
                                    const string &userName, const string &userPassword,
                                    const string &roleName, const string &charSet,
                                    const string &createParams, const string &login, const string &loginPassword, bool isSha1 = true); // throw(...);

        void                   connect(const UnicodeString &serverName, const UnicodeString &databaseName,
                                    const UnicodeString &userName, const UnicodeString &userPassword,
                                    const UnicodeString &roleName, const UnicodeString &charSet,
                                    const UnicodeString &createParams, const UnicodeString &login, const UnicodeString &loginPassword, bool isSha1 = true); // throw(...);

        void                   connect(const string &serverName, const string &databaseName, 
                                    const string &userName, const string &userPassword,
                                    const string &roleName, const string &charSet,
                                    const string &createParams, const string &login, const string &loginPassword, bool isSha1 = true); // throw(...);

        DatabaseModule(const UnicodeString &serverName, const UnicodeString &databaseName,
                                    const UnicodeString &userName, const UnicodeString &userPassword,
                                    const UnicodeString &roleName, const UnicodeString &charSet,
                                    const UnicodeString &createParams); // throw(...);

        DatabaseModule(const string &serverName, const string &databaseName,
                                    const string &userName, const string &userPassword,
                                    const string &roleName, const string &charSet,
                                    const string &createParams); // throw(...);

        void                   connect(const UnicodeString &serverName, const UnicodeString &databaseName,
                                    const UnicodeString &userName, const UnicodeString &userPassword,
                                    const UnicodeString &roleName, const UnicodeString &charSet,
                                    const UnicodeString &createParams); // throw(...);

        void                   connect(const string &serverName, const string &databaseName,
                                    const string &userName, const string &userPassword,
                                    const string &roleName, const string &charSet,
                                    const string &createParams); // throw(...);


        string                 getServerName()
        {
          return this->m_db->ServerName();
        }

        string                 getDatabaseName()
        {
          return this->m_db->DatabaseName();
        }

        string                 getUserName()
        {
          return this->m_db->Username();
        }

        string                 getUserPassword()
        {
          return this->m_db->UserPassword();
        }

        string                 getRoleName()
        {
          return this->m_db->RoleName();
        }

        string                 getCharset()
        {
          return this->m_db->CharSet();
        }

        string                 getCreateParams()
        {
          return this->m_db->CreateParams();
        }

        /**
         * TAM - Transaction Access Mode
         * TIL - Transaction Isolation Level
         *       ilConcurrency -  called Snapshot Repeatable Read (it allows a transaction to get a snapshot 
         *                        of the complete database at the time it's started,
         *                        ideal for reporting, not recommended 
         *                        for interactive users, this transaction cannot see any change made by 
         *                        other concurrent transactions, that transaction only can see changes 
         *                        made by itself)
         *       ilConsistency - called Snapshot table stability aka Forced Repeatable Read
         *                       specific to Firebird/InterBase
         *                       (as soon as it reads from a table, this table is write-locked for all 
         *                       other transactions,
         *                       transaction with this level takes control of that table and all other 
         *                       transactions only can read from it)
         * TLR - Transaction Lock Resolution
         * TFF - TransactionFactory Flags
         */
        Transaction            createTransaction(TAM am = amWrite, TIL il = ilReadCommitted, 
                                                 TLR lr = lrNoWait, TFF flags = TFF(0));// throw(...);

        Statement              createStatement(Transaction *tr, string *sql = (string *) 0);// throw(...);

        void                   checkTableLock(const int64_t tableId, const string &tablePrefix, 
                                              const string &tableName, 
                                              Transaction *tr);
                                              //throw(...);

        int64_t                insertIntoTable(const string &tablePrefix, const string &tableName, 
                                               boost::ptr_vector<DBField> &dbfv, Transaction *tr = (Transaction *) 0);
                                               //throw(...);

        int64_t                insertIntoTable2(const string &tablePrefix, const string &tableName,
                                               boost::ptr_vector<DBField> &dbfv, Transaction *tr = (Transaction *) 0);
                                               //throw(...);

        int64_t                insertIntoTableNA(const string &tablePrefix, const string &tableName,
                                               boost::ptr_vector<DBField> &dbfv, Transaction *tr = (Transaction *) 0);
                                               //throw(...);

        void                   updateTable(const int64_t tableId, const string &tablePrefix, 
                                        const string &tableName, boost::ptr_vector<DBField> &dbfv, 
                                        Transaction *tr);
                                        //throw (...);

        void                   updateTable(const string  &whereSql, const string &tablePrefix, 
                                        const string &tableName, boost::ptr_vector<DBField> &dbfv, 
                                        Transaction *tr);
                                        //throw (...);

        void                   updateTableNA(const int64_t tableId, const string &tablePrefix,
                                        const string &tableName, boost::ptr_vector<DBField> &dbfv,
                                        Transaction *tr);
                                        //throw (...);

        void                   updateTableNA(const string  &whereSql, const string &tablePrefix,
                                        const string &tableName, boost::ptr_vector<DBField> &dbfv,
                                        Transaction *tr);
                                        //throw (...);


        void                   deleteFromTable(const int64_t tableId, const string &tablePrefix, 
                                        const string &tableName, Transaction *tr);
                                        //throw (...);

        void                   deleteFromTable(const string &whereSql, const string &tablePrefix, 
                                        const string &tableName, Transaction *tr);
                                        //throw (...);

        void                   login(const string login, const string password, Transaction *tr, bool isSha1 = true);//throw(...);

        bool                   logged();
        string                 getName();
        string                 getSurname();
        string                 getHashPassword();
        int64_t                getUserID();
        bool                   isUser(int64_t privilage);

        /*
         * W przyszłości updatetować tabele np. MIR_ZALOGOWANI
         **/
        bool                   logout();// throw(...);


        static string          toSql(const string &input, const bool addSingleQuotations = true);

        static string          toSimilarToSql(const string &input, const bool addEscapeWord = true, const bool escapeHypens = false, const bool escapeQuestionMarks = false);

    }; // class DatabaseModule

  } // namespace MIR
} // namespace SyNaT
#endif //_MIR_DATABASEMODULE_H_
