#ifndef _MIR_URI_H_
#define _MIR_URI_H_
/*
 * Copyright (c) 2011,
 * Technical University Of Gdansk,
 * Faculty of Electronics, Telecommunications and Informatics,
 * Multimedia Systems Department
 * and/or its affiliates. All rights reserved.
 */
#include <boost/ptr_container/ptr_vector.hpp>

#include <unicode/unistr.h>

namespace SyNaT
{
  namespace MIR
  {
    using namespace std;

    class IcuURI
    {
      public:
        enum FieldTypes
        {
          IcuURI_SCHEME = 1,
          IcuURI_USERINFO = 2,
          IcuURI_SERVER = 4,
          IcuURI_PORT = 8,
          IcuURI_PATH = 16,
          IcuURI_QUERY = 32,
          IcuURI_FRAGMENT = 64
        };

        enum ResolveTypes
        {
          IcuURI_STRICT = 1
        };

        IcuURI();
        IcuURI(const UnicodeString uri);
        IcuURI(const IcuURI& uri);
        // default copy ctor, assignment operator and dtor are ok

        virtual ~IcuURI()
        {
        }

        bool Create(const UnicodeString &uri);

        IcuURI& operator=(const UnicodeString &string)
        {
            Create(string);
            return *this;
        }

        IcuURI& operator=(const IcuURI& uri)
        {
          this->m_scheme = uri.m_scheme;
          this->m_userinfo = uri.m_userinfo;
          this->m_server = uri.m_server;
          this->m_port = uri.m_port;
          this->m_path = uri.m_path;
          this->m_query = uri.m_query;
          this->m_fragment = uri.m_fragment;
          this->m_fields = uri.m_fields;
          return *this;
        }

        bool operator==(const IcuURI& uri) const;

        // various accessors

        bool HasScheme() const      { return (m_fields & IcuURI_SCHEME) != 0;   }
        bool HasUserInfo() const    { return (m_fields & IcuURI_USERINFO) != 0; }
        bool HasServer() const      { return (m_fields & IcuURI_SERVER) != 0;   }
        bool HasPort() const        { return (m_fields & IcuURI_PORT) != 0;     }
        bool HasPath() const        { return (m_fields & IcuURI_PATH) != 0;     }
        bool HasQuery() const       { return (m_fields & IcuURI_QUERY) != 0;    }
        bool HasFragment() const    { return (m_fields & IcuURI_FRAGMENT) != 0; }

        const UnicodeString& GetScheme() const    { return m_scheme;   }
        const UnicodeString& GetPath() const      { return m_path;     }
        const UnicodeString& GetQuery() const     { return m_query;    }
        const UnicodeString& GetFragment() const  { return m_fragment; }
        const UnicodeString& GetPort() const      { return m_port;     }
        const UnicodeString& GetUserInfo() const  { return m_userinfo; }
        const UnicodeString& GetServer() const    { return m_server;   }

        bool IsReference() const { return !HasScheme() || !HasServer(); }

        // these functions only work if the user information part of the URI is in
        // the usual (but insecure and hence explicitly recommended against by the
        // RFC) "user:password" form
        UnicodeString GetUser() const;
        UnicodeString GetPassword() const;


        // combine all URI components into a single string
        //
        // BuildURI() returns the real URI suitable for use with network libraries,
        // for example, while BuildUnescapedURI() returns a string suitable to be
        // shown to the user.
        UnicodeString BuildURI() const;
        UnicodeString BuildUnescapedURI() const;

        // the escaped URI should contain only ASCII characters, including possible
        // escape sequences
        static UnicodeString Unescape(const UnicodeString escapedURI);

        void SplitInSegments(const UnicodeString &path, boost::ptr_vector<UnicodeString> &result);

        UnicodeString JoinSegments(const boost::ptr_vector<UnicodeString> &input);

        void Resolve(const IcuURI& base, int flags = IcuURI_STRICT);

      protected:

        UnicodeString  m_scheme;
        UnicodeString  m_path;
        UnicodeString  m_query;
        UnicodeString  m_fragment;

        UnicodeString  m_userinfo;
        UnicodeString  m_server;
        UnicodeString  m_port;

        int m_fields;
    };

  }
}

#endif /* _MIR_URI_H_ */
