#include <boost/foreach.hpp>
#include <boost/ptr_container/ptr_vector.hpp>

#include <string>

#include "utils/mir_utils.h"
#include "mir_exceptions.h"
#include "utils/mir_uri.h"

#include <unicode/unistr.h>
#include <unicode/regex.h>

namespace SyNaT
{
  namespace MIR
  {
    using namespace std;

    IcuURI::IcuURI()
      : m_fields(0)
    {               //12            3  4          5       6   7        8 9
                    //example: http://www.ics.uci.edu/pub/ietf/uri/#Related
                    // 1 = "http:"
                    // 2 = "http"
                    // 3 = "//www.ics.uci.edu"
                    // 4 = "www.ics.uci.edu"
                    // 5 = "/pub/ietf/uri/"
                    // 6 = ""
                    // 7 = ""
                    // 8 = "#Related"
                    // 9 = "Related"
		int i = INT8_MIN;
    }

    IcuURI::IcuURI(const UnicodeString uri)
    : m_fields(0)
    {             //12            3  4          5       6   7        8 9
      this->Create(uri);
    }

    IcuURI::IcuURI(const IcuURI& uri)
    : m_fields(0)
    {
      this->m_scheme = uri.m_scheme;
      this->m_userinfo = uri.m_userinfo;
      this->m_server = uri.m_server;
      this->m_port = uri.m_port;
      this->m_path = uri.m_path;
      this->m_query = uri.m_query;
      this->m_fragment = uri.m_fragment;
      this->m_fields = uri.m_fields;
    }

    bool IcuURI::Create(const UnicodeString &uri)
    {
      UErrorCode      status_regex_uri = U_ZERO_ERROR;
      RegexMatcher   *regex_uri = new RegexMatcher(UnicodeString::fromUTF8("^(([^:\\/\\?#]+):\\/\\/)?(([^\\/\\?#]*))?([^\\?#]*)(\\?([^#]*))?(#(.*))?"), 0, status_regex_uri);
      if (U_FAILURE(status_regex_uri))
        throw RegexMatcherMirException("bool IcuURI::Create(const UnicodeString uri)");

      regex_uri->reset(uri);
      bool retVal = false;
      if (regex_uri->find())
      {
        status_regex_uri = U_ZERO_ERROR;
        if (regex_uri->group(1, status_regex_uri).length() > 0)
        {
          this->m_fields |= IcuURI_SCHEME;
        }

        status_regex_uri = U_ZERO_ERROR;
        this->m_scheme = regex_uri->group(2, status_regex_uri);

        status_regex_uri = U_ZERO_ERROR;
        UnicodeString serverWithUserInfoAndPort = regex_uri->group(4, status_regex_uri);
        int32_t lioAT = serverWithUserInfoAndPort.lastIndexOf("@");
        if (lioAT >= 0)
        {
          this->m_userinfo = UnicodeString(serverWithUserInfoAndPort, 0, lioAT);
          this->m_fields |= IcuURI_USERINFO;
        }
        int32_t lioColon = serverWithUserInfoAndPort.lastIndexOf(":");
        if ((lioColon > 0) && (lioColon > lioAT))
        {
          this->m_port =  UnicodeString(serverWithUserInfoAndPort, lioColon + 1);
          this->m_fields |= IcuURI_PORT;
        }

        if(lioColon > lioAT)
        {
          this->m_server = UnicodeString(serverWithUserInfoAndPort, lioAT + 1, lioColon - lioAT - 1);
          this->m_fields |= IcuURI_SERVER;
        }
        else if (lioColon < lioAT)
        {
          this->m_server = UnicodeString(serverWithUserInfoAndPort, lioAT + 1);
          this->m_fields |= IcuURI_SERVER;
        }
        else
        {
          this->m_server = serverWithUserInfoAndPort;
          if (!this->m_server.isEmpty())
            this->m_fields |= IcuURI_SERVER;
        }

        status_regex_uri = U_ZERO_ERROR;
        this->m_path = regex_uri->group(5, status_regex_uri);
        if (!this->m_path.isEmpty())
          this->m_fields |= IcuURI_PATH;

        status_regex_uri = U_ZERO_ERROR;
        if (!regex_uri->group(6, status_regex_uri).isEmpty())
          this->m_fields |= IcuURI_QUERY;

        status_regex_uri = U_ZERO_ERROR;
        this->m_query = regex_uri->group(7, status_regex_uri);

        status_regex_uri = U_ZERO_ERROR;
        if (!regex_uri->group(8, status_regex_uri).isEmpty())
          this->m_fields |= IcuURI_FRAGMENT;

        status_regex_uri = U_ZERO_ERROR;
        this->m_fragment = regex_uri->group(9, status_regex_uri);
        retVal = true;
      }

      delete regex_uri;
      return retVal;
    }

    bool IcuURI::operator ==(const IcuURI & uri) const
    {
      if (this->HasScheme() && uri.HasScheme() && this->m_scheme.compare(uri.m_scheme))
        return false;
      if (this->HasUserInfo() && uri.HasUserInfo() && this->m_userinfo.compare(uri.m_userinfo))
        return false;
      if (this->HasServer() && uri.HasServer() && this->m_server.compare(uri.m_server))
        return false;
      if (this->HasPort() && uri.HasPort() && this->m_port.compare(uri.m_port))
        return false;
      if (this->HasPath() && uri.HasPath() && this->m_path.compare(uri.m_path))
        return false;
      if (this->HasQuery() && uri.HasQuery() && this->m_query.compare(uri.m_query))
        return false;
      if (this->HasFragment() && uri.HasFragment() && this->m_fragment.compare(uri.m_fragment))
        return false;

      return true;
    }

    UnicodeString IcuURI::GetUser() const
    {
      return this->HasUserInfo() ?
          UnicodeString(this->m_userinfo, 0, this->m_userinfo.lastIndexOf(":"))
          : "";
    }

    UnicodeString IcuURI::GetPassword() const
    {
      return this->HasUserInfo() ?
          UnicodeString(this->m_userinfo, this->m_userinfo.lastIndexOf(":") + 1)
          : "";
    }

    UnicodeString IcuURI::BuildURI() const
    {
      UnicodeString uri;
      if (this->HasScheme())
        uri += this->m_scheme + UnicodeString::fromUTF8("://");
      if (this->HasUserInfo())
        uri += this->m_userinfo + '@';
      if (this->HasServer())
        uri += this->m_server;
      if (this->HasPort())
        uri += ':' + this->m_port;
      if (this->HasPath())
        uri += Utils::encode(this->m_path);
      if (this->HasQuery())
        uri += UnicodeString::fromUTF8("?") + Utils::encode(this->m_query);
      if (this->HasFragment())
        uri += UnicodeString::fromUTF8("#") + this->m_fragment;
      return uri;
    }

    UnicodeString IcuURI::BuildUnescapedURI() const
    {
      UnicodeString uri;
      if (this->HasScheme())
        uri += this->m_scheme;
      if (this->HasUserInfo())
        uri += this->m_userinfo + '@';
      if (this->HasServer())
        uri += this->m_server;
      if (this->HasPort())
        uri += ':' + this->m_port;
      if (this->HasPath())
        uri += Utils::encode(this->m_path);
      if (this->HasQuery())
        uri += '?' + this->m_query;
      if (this->HasFragment())
        uri += '#' + this->m_fragment;
      return uri;
    }

    UnicodeString IcuURI::Unescape(const UnicodeString escapedURI)
    {
      IcuURI uri(escapedURI);
      return uri.BuildUnescapedURI();
    }
/*

  void wxURI::Resolve(const wxURI& base, int flags)
  {
      wxASSERT_MSG(!base.IsReference(),
                  "wxURI to inherit from must not be a reference!");

      // If we aren't being strict, enable the older (pre-RFC2396) loophole that
      // allows this uri to inherit other properties from the base uri - even if
      // the scheme is defined
      if ( !(flags & wxURI_STRICT) &&
              HasScheme() && base.HasScheme() &&
                  m_scheme == base.m_scheme )
      {
          m_fields -= wxURI_SCHEME;
      }


      // Do nothing if this is an absolute wxURI
      //    if defined(R.scheme) then
      //       T.scheme    = R.scheme;
      //       T.authority = R.authority;
      //       T.path      = remove_dot_segments(R.path);
      //       T.query     = R.query;
      if (HasScheme())
          return;

      //No scheme - inherit
      m_scheme = base.m_scheme;
      m_fields |= wxURI_SCHEME;

      // All we need to do for relative URIs with an
      // authority component is just inherit the scheme
      //       if defined(R.authority) then
      //          T.authority = R.authority;
      //          T.path      = remove_dot_segments(R.path);
      //          T.query     = R.query;
      if (HasServer())
          return;

      //No authority - inherit
      if (base.HasUserInfo())
      {
          m_userinfo = base.m_userinfo;
          m_fields |= wxURI_USERINFO;
      }

      m_server = base.m_server;
      m_hostType = base.m_hostType;
      m_fields |= wxURI_SERVER;

      if (base.HasPort())
      {
          m_port = base.m_port;
          m_fields |= wxURI_PORT;
      }


      // Simple path inheritance from base
      if (!HasPath())
      {
          //             T.path = Base.path;
          m_path = base.m_path;
          m_fields |= wxURI_PATH;


          //             if defined(R.query) then
          //                T.query = R.query;
          //             else
          //                T.query = Base.query;
          //             endif;
          if (!HasQuery())
          {
              m_query = base.m_query;
              m_fields |= wxURI_QUERY;
          }
      }
      else if ( m_path.empty() || m_path[0u] != '/' )
      {
          //             if (R.path starts-with "/") then
          //                T.path = remove_dot_segments(R.path);
          //             else
          //                T.path = merge(Base.path, R.path);
          //                T.path = remove_dot_segments(T.path);
          //             endif;
          //             T.query = R.query;
          //
          // So we don't do anything for absolute paths and implement merge for
          // the relative ones

          wxArrayString our(SplitInSegments(m_path)),
                        result(SplitInSegments(base.m_path));

          if ( !result.empty() )
              result.pop_back();

          if ( our.empty() )
          {
              // if we have an empty path it means we were constructed from a "."
              // string or something similar (e.g. "././././"), it should count
              // as (empty) segment
              our.push_back("");
          }

          const wxArrayString::const_iterator end = our.end();
          for ( wxArrayString::const_iterator i = our.begin(); i != end; ++i )
          {
              if ( i->empty() || *i == "." )
              {
                  // as in ParsePath(), while normally we ignore the empty
                  // segments, we need to take account of them at the end
                  if ( i == end - 1 )
                      result.push_back("");
                  continue;
              }

              if ( *i == ".." )
              {
                  if ( !result.empty() )
                  {
                      result.pop_back();

                      if ( i == end - 1 )
                          result.push_back("");
                  }
                  //else: just ignore, extra ".." don't accumulate
              }
              else
              {
                  if ( result.empty() )
                  {
                      // ensure that the resulting path will always be absolute
                      result.push_back("");
                  }

                  result.push_back(*i);
              }
          }

          m_path = wxJoin(result, '/', '\0');
      }

      //T.fragment = R.fragment;
  }
*/
    void IcuURI::SplitInSegments(const UnicodeString &path, boost::ptr_vector<UnicodeString> &result)
    {
      UErrorCode status = U_ZERO_ERROR;
      RegexMatcher *m = new RegexMatcher("\\/([^\\/]*)", 0, status);
      m->reset(path);
      while(m->find())
      {
        result.push_back(new UnicodeString(m->group(1, status)));
      }
      delete m;
    }

    UnicodeString IcuURI::JoinSegments(const boost::ptr_vector<UnicodeString> &input)
    {
      UnicodeString us;
      BOOST_FOREACH(const UnicodeString &uspart, input)
      {
        us += "/" + uspart;
      }
      return us;
    }

    void IcuURI::Resolve(const IcuURI & base, int flags)
    {
      if (base.IsReference())
        throw BaseURIIsAReferenceMirException("void IcuURI::Resolve(const IcuURI & base, int flags)");

      if ( !(flags & IcuURI_STRICT) &&
              HasScheme() && base.HasScheme() &&
                  m_scheme == base.m_scheme )
      {
          m_fields -= IcuURI_SCHEME;
      }

      if (HasScheme())
          return;

      m_scheme = base.m_scheme;
      m_fields |= IcuURI_SCHEME;

      if (base.HasUserInfo())
      {
          m_userinfo = base.m_userinfo;
          m_fields |= IcuURI_USERINFO;
      }

      m_server = base.m_server;
      m_fields |= IcuURI_SERVER;

      if (base.HasPort())
      {
          m_port = base.m_port;
          m_fields |= IcuURI_PORT;
      }

      if (!HasPath())
      {
          m_path = base.m_path;
          m_fields |= IcuURI_PATH;

          if (!HasQuery())
          {
              m_query = base.m_query;
              m_fields |= IcuURI_QUERY;
          }
      }
      else if ( m_path.isEmpty() || m_path[0u] != '/' )
      {
        //             if (R.path starts-with "/") then
        //                T.path = remove_dot_segments(R.path);
        //             else
        //                T.path = merge(Base.path, R.path);
        //                T.path = remove_dot_segments(T.path);
        //             endif;
        //             T.query = R.query;
        //
        // So we don't do anything for absolute paths and implement merge for
        // the relative ones
        boost::ptr_vector<UnicodeString> our, result;
        this->SplitInSegments(m_path, our);
        this->SplitInSegments(base.m_path, result);

        if ( !result.empty() )
            result.pop_back();

        if ( our.empty() )
        {
            // if we have an empty path it means we were constructed from a "."
            // string or something similar (e.g. "././././"), it should count
            // as (empty) segment
            our.push_back(new UnicodeString(""));
        }

        const boost::ptr_vector<UnicodeString>::const_iterator end = our.end();
        for ( boost::ptr_vector<UnicodeString>::const_iterator i = our.begin(); i != end; ++i )
        {
          if ( i->isEmpty() || *i == "." )
          {
            // as in ParsePath(), while normally we ignore the empty
            // segments, we need to take account of them at the end
            if ( i == end - 1 )
              result.push_back(new UnicodeString(""));
            continue;
          }

          if ( *i == ".." )
          {
            if ( !result.empty() )
            {
                result.pop_back();

                if ( i == end - 1 )
                    result.push_back(new UnicodeString(""));
            }
            //else: just ignore, extra ".." don't accumulate
          }
          else
          {
            if ( result.empty() )
            {
                // ensure that the resulting path will always be absolute
                result.push_back(new UnicodeString(""));
            }

            result.push_back(new UnicodeString(*i));
          }
        }

        m_path = this->JoinSegments(result);
      }

    }

  }
}
